#!/usr/bin/perl -w
########################################################################
# File: perqd
# Abstracts:
#	  This perl script is a daemon for managing queues of files
# History:
# Ver 1.00 Original
#		by maoya@mail.goo.ne.jp in Sun May 26 12:27:52 JST 2002
########################################################################

## Several Declarations: Perl version, Strict, Socket, Directory Handler
require 5.002;
use strict;
use Socket;
use DirHandle;

## Switch for debugging
my $DEBUG	= 1;

## Set top directory of the queues
my $top_of_queue	= ".";

## Define the default TCP/IP port number
my $DEFAULT_PORT	= 9999;

## Define the maximum sequence number of queue elements
my $MAX_QUEUE_SEQ	= 9999;

## Define the reply codes
my $NORMAL_END			= "000";
my $ILLEGAL_COMMAND		= "001";
my $SHUTDOWN_ACK		= "002";
my $ENQ_LOST_QUEUE_KEY		= "100";
my $ENQ_NO_EXIST_QUEUE		= "101";
my $ENQ_LOST_SIZE		= "102";
my $ENQ_ILLEGALSIZE		= "103";
my $ENQ_RETRY_OVER_SEQ		= "104";
my $ENDENQ_LOST_QUEUE_ID	= "200";
my $ENDENQ_NO_EXIST_QUEUE_ID	= "201";
my $DEQ_LOST_QUEUE_KEY		= "300";
my $DEQ_NO_EXIST_QUEUE		= "301";
my $DEQ_NO_ELEMENT		= "303";
my $ENDDEQ_LOST_QUEUE_ID	= "400";
my $ENDDEQ_NO_EXIST_QUEUE_ID	= "401";
my $CANENQ_LOST_QUEUE_ID	= "500";
my $CANENQ_NO_EXIST_QUEUE_ID	= "501";
my $CANDEQ_LOST_QUEUE_ID	= "600";
my $CANDEQ_NO_EXIST_QUEUE_ID	= "601";
my $ENQ_HEAD_LOST_QUEUE_KEY	= "701";
my $ENQ_HEAD_NO_EXIST_QUEUE	= "702";
my $ENQ_HEAD_LOST_SIZE		= "703";
my $ENQ_HEAD_ILLEGALSIZE	= "704";

########################################################################
# Start Procedure
########################################################################
## Variables for accepting to connect requests from client
my ($addr, $proto, $port, $iaddr);

## Set the number of listening port
if (exists $ENV{PERQ_PORT} && $ENV{PERQ_PORT} ne "") {
	$port	= $ENV{PERQ_PORT}; # Listening to the specified port
}
else {
	$port	= $DEFAULT_PORT; # Listening to the default port
}

## Set the protocol (tcp/ip)
$proto	= getprotobyname('tcp');

## Create the socket of the internet domain (reusable address option)
socket(S, PF_INET, SOCK_STREAM, $proto)
	or die "socket($port): $!";
setsockopt(S, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
	or die "setsockopt(SO_REUSEADDR): $!";

##   Bind the socket to the port listening to connect requests
## from any hosts
$addr	= sockaddr_in($port, inet_aton("localhost"));
#bind(S, sockaddr_in($port, INADDR_ANY))
bind(S, $addr) or die "bind($port, INADDR_ANY): $!";
listen(S, SOMAXCONN) or die "listen(): $!";

## Wait for the connect request
for (; $addr = accept(CL, S); close(CL)) {
	($port, $iaddr)	= sockaddr_in($addr);
	my ($line, $cmd, $qkey, $size, $now, $dir, $dh, $id, $seq,
		$resp, $ii);
	$resp	= "";

	## Read a line from client, and get the current time (secs)
	chomp($line = <CL>);

	## Check the command
	if ($line =~ /^ENQ-HEAD\s/) {
		## Enqueue (Insert to head of queue) command: 
		##	ENQ-HEAD qkwy size
		my ($ss);
		($cmd, $qkey, $size)	= split /\s/, $line;

		## Cheak arguments
		if ($qkey eq "") {
			## Enqueue (Inset to head of queue) must have
			## queue-key argument
			print CL $ENQ_HEAD_LOST_QUEUE_KEY .
				" Lost Queue Key\n";	# Error
			next;
		}
		elsif (! -d "$top_of_queue/$qkey") {
			## The directory as queue key is not exist
			print CL $ENQ_HEAD_NO_EXIST_QUEUE .
				" No-exist Queue Key ($qkey)\n"; # Error
			next;
		}
		if ($size eq "") {
			## Enqueue command must have size argument
			print CL $ENQ_HEAD_LOST_SIZE .
				" Lost Element Size\n";	# Error
			next;
		}
		elsif ($size < 0) {
			## Enqueue command must have size
			print CL $ENQ_HEAD_ILLEGALSIZE .
				" Illegal Size ($size)\n"; # Error
			next;
		}

		## Create directory handle as queue key
		($dh	= new DirHandle("$top_of_queue/$qkey"))
			or die "DirHandle($top_of_queue/$qkey): $!\n";

		$now	= time();	# Get current time
		$ii = $id	= $now;
		$ss = $seq	= $MAX_QUEUE_SEQ;
		while ($_ = $dh->read) {
			$_	=~ tr/\+//d;
			$_	=~ tr/\-//d;
			($ii, $ss)	= split /\./, $_;
			if ($ii eq "") {
				next;	# dot-file are skipped
			}
			if ($ii eq "" || $ss >= $MAX_QUEUE_SEQ) {
				last; # "seq" must be 4 digits
			}
			if ($ii < $id) {
				$id	= $ii;
				$seq	= $ss;
			}
			elsif ($ii == $id && $ss < $seq) {
				$seq	= $ss;
			}
		}
		$dh->close;

		if ($seq == 0) {
			--$id;
			$seq	= $MAX_QUEUE_SEQ;
		}
		else {
			--$seq;
		}

		## Get the response (path of queue element)
		$resp	= sprintf("%s/%s/+%s.%04d",
				$top_of_queue, $qkey, $id, $seq);

		## Create the empty file as queue element
		open(QE, ">$resp"); print QE ""; close(QE);
	}
	elsif ($line =~ /^ENQ\s/) {
		## Enqueue command: ENQ qkey size
		my ($ss);

		$now	= time();
		($cmd, $qkey, $size)	= split /\s/, $line;

		## Check arguments
		if ($qkey eq "") {
			## Enqueue command must have queue-key argument
			print CL $ENQ_LOST_QUEUE_KEY .
				" Lost Queue Key\n"; # Error
			next;
		}
		elsif (! -d "$top_of_queue/$qkey") {
			## The directory as queue key is not exist
			print CL $ENQ_NO_EXIST_QUEUE .
				" No-exist Queue Key ($qkey)\n"; # Error
			next;
		}
		if ($size eq "") {
			## Enqueue command must have size argument
			print CL $ENQ_LOST_SIZE .
				" Lost Element Size\n";	# Error
			next;
		}
		elsif ($size < 0) {
			## Enqueue command must have size
			print CL $ENQ_ILLEGALSIZE .
				" Illegal Size ($size)\n"; # Error
			next;
		}

		## Create directory handle as queue key
		($dh	= new DirHandle("$top_of_queue/$qkey"))
			or die "DirHandle($top_of_queue/$qkey): $!\n";

		## Get the sequential number for queue element
		$id	= $now;
		$ss = $seq	= -1;
		while ($_ = $dh->read) {
			if ($_ =~ /^\+$now\./) {
				($id, $ss)	= split /\./, $_;
				if ($ss >= $MAX_QUEUE_SEQ) {
					last; # "seq" must be 4 digits
				}
				if ($ss > $seq) {
					$seq	= $ss;
				}
			}
		}
		$dh->close;

		##   If the sequential number is over 4 digits,
		## reply "Retry" message to client 
		if ($ss >= $MAX_QUEUE_SEQ) {
			print CL $ENQ_RETRY_OVER_SEQ .
				" Retry (over seq:$ss)\n"; # Retry after
			next;
		}
		$seq++;

		## Get the response (path of queue element)
		$resp	= sprintf("%s/%s/+%s.%04d",
				$top_of_queue, $qkey, $now, $seq);

		## Create the empty file as queue element
		open(QE, ">$resp"); print QE ""; close(QE);
	}
	elsif ($line =~ /^CAN-ENQ\s/) {
		## Cancel Enqueue: CAN-ENQ queue-id

		## Get queue id
		($cmd, $qkey)	= split /\s/, $line;

		## Check queue id
		if ($qkey eq "") {
			print CL $CANENQ_LOST_QUEUE_ID .
				" Lost Queue ID\n";	# Error
			next;
		}
		if (! -f $qkey) {
			print CL $CANENQ_NO_EXIST_QUEUE_ID .
				" No Exist Queue ID ($qkey)\n";	# Error
			next;
		}

		## Remove the queue element
		unlink($qkey);

		$resp	= $qkey;
	}
	elsif ($line =~ /^END-ENQ\s/) {
		## Notice End of Enqueue: END-ENQ queue-id
		my ($newqid);

		## Get queue id
		($cmd, $qkey)	= split /\s/, $line;

		## Check queue id
		if ($qkey eq "") {
			print CL $ENDENQ_LOST_QUEUE_ID .
				" Lost Queue ID\n";
			next;
		}
		if (! -f $qkey) {
			print CL $ENDENQ_NO_EXIST_QUEUE_ID .
				" No Exist Queue-ID ($qkey)\n";
			next;
		}

		## Rename the enqueued element
		$newqid	= $qkey;
		$newqid	=~ tr/\+//d;
		rename $qkey, $newqid;

		$resp	= $newqid;
	}
	elsif ($line =~ /^DEQ\s/) {
		## Dequeue command: DEQ queue-key
		my ($tt, $ss);

		## Get queue key
		($cmd, $qkey)	= split /\s/, $line;

		## Check queue key
		if ($qkey eq "") {
			print CL $DEQ_LOST_QUEUE_KEY .
				" Lost Queue Key\n"; # Error
			next;
		}
		if ( ! -d "$top_of_queue/$qkey") {
			print CL $DEQ_NO_EXIST_QUEUE .
				" No Exist Queue ($qkey)\n"; # Error
			next;
		}

		##   Create directory handle for searching for
		## the earliest queue element.
		($dh	= new DirHandle ("$top_of_queue/$qkey"))
			or die "DirHandle($top_of_queue/$qkey): $!";
		$tt = $now	= 0;
		$ss = $seq	= 10000;
		while ($_ = $dh->read) {
			if ($_ !~ /^[0-9]/) {
				next; # Skip not-queued elements
			}

			($tt, $ss)	= split /\./,$_;
			if ($now == 0 || $tt < $now) {
				$now	= $tt;
				$seq	= $ss;
			}
			elsif ($tt == $now && $ss < $seq) {
				$seq	= $ss;
			}
		}
		$dh->close;
		if ($now == 0) {
			## Reply "No Element" to client
			print CL "$DEQ_NO_ELEMENT No Element to DEQ\n";
			next;
		}

		$id	= "$top_of_queue/$qkey/$now.$seq";
		$resp	= "$top_of_queue/$qkey/-$now.$seq";
		rename($id, $resp);
	}
	elsif ($line =~ /^CAN-DEQ\s/) {
		## Cancel Dequeue: CAN-DEQ queue-id
		my ($element_path);

		## Get queue id
		($cmd, $qkey)	= split /\s/, $line;

		if ($qkey eq "") {
			print CL $CANDEQ_LOST_QUEUE_ID .
				" Lost Queue ID\n";	# Error
			next;
		}
		if (! -f $qkey) {
			print CL $CANDEQ_NO_EXIST_QUEUE_ID .
				" No Exist Queue ID ($qkey)\n";	# Error
			next;
		}

		$element_path	= $qkey;
		$element_path	=~ tr/-//d;
		rename	$qkey, $element_path;

		$resp	= $element_path;
	}
	elsif ($line =~ /^END-DEQ\s/) {
		## Notice End of Dequeue: END-DEQ queue-id

		## Get queue id
		($cmd, $qkey)	= split /\s/,$line;

		if ($qkey eq "") {
			print CL $ENDDEQ_LOST_QUEUE_ID .
				" Lost Queue ID\n";	# Error
			next;
		}
		if (! -f "$qkey") {
			print CL $ENDDEQ_NO_EXIST_QUEUE_ID .
				" No Exist Queue ID ($qkey)\n"; # Error
			next;
		}

		unlink $qkey;
		$resp	= $qkey;
	}
	elsif ($line =~ /^SHUTDOWN\s/) {
		## Shut-down command: SHUTDOWN
		print CL "$SHUTDOWN_ACK Server Going Down...Bye\!\n";
		last;	# Break the listening loop
	}
	else {	## Received unknown command
		print CL $ILLEGAL_COMMAND .
			" Illegal Command\n"; # Error
		next;
	}

	## Reply to the client (Normal End)
	print CL "000 $resp\n";
}

## Finish
shutdown S, 3; close S;

exit 0;	# Stop this daemon


